"""Tests for semantic tokens provider."""

from textwrap import dedent

from collagraph_lsp.semantic_tokens import SemanticTokensProvider


def test_semantic_tokens_provider_initialization():
    """Test that the provider initializes correctly."""
    provider = SemanticTokensProvider()
    assert len(provider.token_types) > 0
    assert len(provider.token_modifiers) > 0
    assert "class" in provider.token_types
    assert "function" in provider.token_types
    assert "declaration" in provider.token_modifiers


def test_get_tokens_from_script():
    """Test extracting tokens from script section."""
    content = dedent(
        """
        <template>
            <widget />
        </template>

        <script>
        import collagraph as cg

        class MyComponent(cg.Component):
            def init(self):
                self.state["count"] = 0

            def increment(self, ev):
                self.state["count"] += 1
        </script>
        """
    ).lstrip()
    provider = SemanticTokensProvider()
    tokens = provider.get_tokens(content)

    # Should find class and method definitions
    token_types = [t.token_type for t in tokens]
    assert "class" in token_types
    assert "method" in token_types or "function" in token_types
    assert "parameter" in token_types


def test_get_tokens_from_template():
    """Test extracting tokens from template section."""
    content = dedent(
        """
        <template>
            <widget>
                <button @clicked="increment" :text="label" />
            </widget>
        </template>

        <script>
        class MyComponent:
            def increment(self):
                pass
        </script>
        """
    ).lstrip()
    provider = SemanticTokensProvider()
    tokens = provider.get_tokens(content)

    # Should find template elements and directives
    token_types = [t.token_type for t in tokens]
    assert "class" in token_types  # widget/button tags
    # Note: directives and events are harder to test without exact positions


def test_get_tokens_empty_file():
    """Test handling of empty file."""
    content = ""
    provider = SemanticTokensProvider()
    tokens = provider.get_tokens(content)
    assert tokens == []


def test_get_tokens_no_script():
    """Test handling file with no script section."""
    content = dedent(
        """
        <template>
            <widget />
        </template>
        """
    ).lstrip()
    provider = SemanticTokensProvider()
    tokens = provider.get_tokens(content)
    # Should still extract template tokens
    assert isinstance(tokens, list)


def test_get_tokens_invalid_python():
    """Test handling invalid Python in script section."""
    content = dedent(
        """
        <template>
            <widget />
        </template>

        <script>
        def invalid syntax here
        </script>
        """
    ).lstrip()
    provider = SemanticTokensProvider()
    tokens = provider.get_tokens(content)
    # Should not crash, just return empty or partial tokens
    assert isinstance(tokens, list)


def test_encode_tokens():
    """Test encoding tokens into LSP format."""
    provider = SemanticTokensProvider()

    # Create sample tokens
    from collagraph_lsp.semantic_tokens import SemanticToken

    tokens = [
        SemanticToken(
            line=0,
            start_char=0,
            length=5,
            token_type="class",
            modifiers=["declaration"],
        ),
        SemanticToken(
            line=0, start_char=6, length=3, token_type="function", modifiers=[]
        ),
        SemanticToken(
            line=1, start_char=4, length=4, token_type="variable", modifiers=[]
        ),
    ]

    encoded = provider.encode_tokens(tokens)

    # Should be a flat list of integers (5 per token)
    assert len(encoded) == len(tokens) * 5
    assert all(isinstance(x, int) for x in encoded)

    # First token
    assert encoded[0] == 0  # delta line
    assert encoded[1] == 0  # delta start
    assert encoded[2] == 5  # length

    # Second token (same line)
    assert encoded[5] == 0  # delta line (same line)
    assert encoded[6] == 6  # delta start (from previous)
    assert encoded[7] == 3  # length

    # Third token (next line)
    assert encoded[10] == 1  # delta line
    assert encoded[11] == 4  # delta start (absolute since new line)
    assert encoded[12] == 4  # length


def test_encode_tokens_empty():
    """Test encoding empty token list."""
    provider = SemanticTokensProvider()
    encoded = provider.encode_tokens([])
    assert encoded == []


def test_class_detection():
    """Test that class definitions are properly detected."""
    content = dedent(
        """
        <script>
        class Foo:
            pass

        class Bar(Foo):
            pass
        </script>
        """
    ).lstrip()
    provider = SemanticTokensProvider()
    tokens = provider.get_tokens(content)

    class_tokens = [t for t in tokens if t.token_type == "class"]
    assert len(class_tokens) >= 2  # Should find both Foo and Bar


def test_method_detection():
    """Test that methods are properly detected."""
    content = dedent(
        """
        <script>
        class MyClass:
            def method_one(self):
                pass

            def method_two(self, param):
                pass
        </script>
        """
    ).lstrip()
    provider = SemanticTokensProvider()
    tokens = provider.get_tokens(content)

    method_tokens = [t for t in tokens if t.token_type == "method"]
    assert len(method_tokens) >= 2  # Should find both methods


def test_parameter_detection():
    """Test that parameters are properly detected (excluding self)."""
    content = dedent(
        """
        <script>
        def my_function(arg1, arg2, arg3):
            pass

        class MyClass:
            def method(self, param1, param2):
                pass
        </script>
        """
    ).lstrip()
    provider = SemanticTokensProvider()
    tokens = provider.get_tokens(content)

    param_tokens = [t for t in tokens if t.token_type == "parameter"]
    # Should find: arg1, arg2, arg3, param1, param2 (not self)
    assert len(param_tokens) >= 5


def test_import_detection():
    """Test that imports are properly detected."""
    content = dedent(
        """
        <script>
        import collagraph
        from collagraph import Component
        from typing import List, Dict
        </script>
        """
    ).lstrip()
    provider = SemanticTokensProvider()
    tokens = provider.get_tokens(content)

    namespace_tokens = [t for t in tokens if t.token_type == "namespace"]
    assert len(namespace_tokens) >= 1  # Should find at least the modules


def test_complex_cgx_file():
    """Test with a complete CGX file similar to real usage."""
    content = dedent(
        """
        <template>
          <widget>
            <button @clicked="increment" :text="button_label" />
            <label :text="counter_text" />
          </widget>
        </template>

        <script>
        import collagraph as cg
        from typing import Dict

        class Counter(cg.Component):
            def __init__(self, props: Dict):
                super().__init__(props)
                self.state["count"] = 0
                self.button_label = "Click me"

            def increment(self):
                self.state["count"] += 1

            @property
            def counter_text(self):
                return f"Count: {self.state['count']}"
        </script>
        """
    ).lstrip()
    provider = SemanticTokensProvider()
    tokens = provider.get_tokens(content)

    # Should find various token types
    token_types = set(t.token_type for t in tokens)
    assert "class" in token_types
    assert "method" in token_types or "function" in token_types
    assert "parameter" in token_types

    # Should be able to encode without errors
    encoded = provider.encode_tokens(tokens)
    assert len(encoded) > 0
    assert len(encoded) % 5 == 0  # Must be multiple of 5



"""Integration tests for the LSP server."""

import textwrap
from importlib.metadata import version

import pytest

from collagraph_lsp.server import (
    CollagraphLanguageServer,
    lint_cgx_content,
    server,
    validate_document,
)


@pytest.fixture
def ls():
    """Fixture providing a fresh language server instance."""
    return CollagraphLanguageServer("collagraph-lsp-test", "v0.1.0")


def test_server_initialization(ls):
    """Test that the server initializes correctly."""
    assert ls.name == "collagraph-lsp-test"
    assert ls.version == "v0.1.0"


def test_server_module_instance():
    """Test that the module-level server instance is properly configured."""
    assert server.name == "collagraph-lsp"
    assert server.version == f"v{version('collagraph_lsp')}"


def test_did_open_handler():
    """Test that the server has a did_open handler registered."""
    from collagraph_lsp.server import did_open

    assert callable(did_open)
    # Check that the handler is decorated with TEXT_DOCUMENT_DID_OPEN
    assert hasattr(did_open, "__name__")


def test_did_change_handler():
    """Test that the server has a did_change handler registered."""
    from collagraph_lsp.server import did_change

    assert callable(did_change)


def test_did_save_handler():
    """Test that the server has a did_save handler registered."""
    from collagraph_lsp.server import did_save

    assert callable(did_save)


def test_did_close_handler():
    """Test that the server has a did_close handler registered."""
    from collagraph_lsp.server import did_close

    assert callable(did_close)


def test_validate_document_with_valid_cgx(ls):
    """Test validation of a valid CGX document."""
    # Test the linter directly rather than full validation
    # (which requires initialized workspace)
    content = textwrap.dedent(
        """
        <template>
          <label :text="message"></label>
        </template>

        <script>
        from collagraph import Component

        class TestComponent(Component):
            message = "Hello"
        </script>
        """
    ).lstrip()

    diagnostics = lint_cgx_content(content)

    # Valid CGX should have no diagnostics (or only expected ones)
    assert isinstance(diagnostics, list)


def test_validate_document_with_python_errors(ls):
    """Test validation of a CGX document with Python errors."""
    content = textwrap.dedent(
        """
        <template>
          <label :text="message"></label>
        </template>

        <script>
        from collagraph import Component
        import unused_module

        class TestComponent(Component):
            undefined_variable = "test"
        </script>
        """
    ).lstrip()

    diagnostics = lint_cgx_content(content)

    # Should detect Python errors
    assert len(diagnostics) > 0


def test_severity_conversion():
    """Test severity string to LSP severity conversion."""
    from lsprotocol.types import DiagnosticSeverity

    from collagraph_lsp.server import _severity_to_lsp

    assert _severity_to_lsp("error") == DiagnosticSeverity.Error
    assert _severity_to_lsp("warning") == DiagnosticSeverity.Warning
    assert _severity_to_lsp("info") == DiagnosticSeverity.Information
    assert _severity_to_lsp("hint") == DiagnosticSeverity.Hint

    # Unknown severity should default to Warning
    assert _severity_to_lsp("unknown") == DiagnosticSeverity.Warning


def test_non_cgx_file_skipped(ls, caplog):
    """Test that validation logic is CGX-specific."""
    import logging

    # Test that non-.cgx files are skipped
    uri = "file:///test.py"

    with caplog.at_level(logging.DEBUG):
        # This should skip validation without raising an exception
        validate_document(ls, uri)

    # Check that the skip message was logged
    assert any("Skipping non-CGX file" in record.message for record in caplog.records)


def test_linter_integration(ls):
    """Test that the server's linter works correctly."""
    test_content = textwrap.dedent(
        """
        <template>
          <label :text="message"></label>
        </template>

        <script>
        from collagraph import Component

        class TestComponent(Component):
            message = "Hello"
        </script>
        """
    ).lstrip()

    diagnostics = lint_cgx_content(test_content)

    # Valid CGX should have no diagnostics (or only expected ones)
    assert isinstance(diagnostics, list)


def test_linter_detects_unused_imports(ls):
    """Test that the linter detects unused imports."""
    test_content = textwrap.dedent(
        """
        <template>
          <label :text="message"></label>
        </template>

        <script>
        from collagraph import Component
        import unused_module

        class TestComponent(Component):
            message = "Hello"
        </script>
        """
    ).lstrip()

    diagnostics = lint_cgx_content(test_content)

    # Should detect the unused import
    assert len(diagnostics) > 0
    assert any(
        "unused_module" in diag.message.lower() or "unused" in diag.message.lower()
        for diag in diagnostics
    )


def test_formatting_handler():
    """Test that the server has a formatting handler registered."""
    from collagraph_lsp.server import formatting

    assert callable(formatting)


def test_format_cgx_content(ls):
    """Test formatting of CGX content through the formatter."""
    from ruff_cgx import format_cgx_content

    content = textwrap.dedent(
        """
        <template>
          <item />
        </template>

        <script>
        from collagraph import   (Component,)
        class   Simple( Component  ):
            pass
        </script>
        """
    ).lstrip()

    formatted = format_cgx_content(content, "test.cgx")

    expected = textwrap.dedent(
        """
        <template>
          <item />
        </template>

        <script>
        from collagraph import (
            Component,
        )


        class Simple(Component):
            pass
        </script>
        """
    ).lstrip()

    assert formatted == expected


def test_initialize_handler_with_ruff_command(ls):
    """Test that INITIALIZE handler accepts and applies ruff_command."""
    from lsprotocol.types import InitializeParams
    from ruff_cgx import get_ruff_command, reset_ruff_command

    from collagraph_lsp.server import initialize

    # Reset to default first
    reset_ruff_command()

    # Create initialization params with ruff_command
    params = InitializeParams(
        process_id=1234,
        root_uri="file:///test",
        capabilities={},
        initialization_options={"ruff_command": "custom-ruff"},
    )

    # Call initialize handler
    initialize(ls, params)

    # Check that settings were stored
    assert ls.settings["ruff_command"] == "custom-ruff"

    # Check that ruff command was applied
    assert get_ruff_command() == "custom-ruff"

    # Reset after test
    reset_ruff_command()


def test_initialize_handler_without_options(ls):
    """Test that INITIALIZE handler works without initialization options."""
    from lsprotocol.types import InitializeParams
    from ruff_cgx import reset_ruff_command

    from collagraph_lsp.server import initialize

    # Reset to default first
    reset_ruff_command()

    # Create initialization params without options
    params = InitializeParams(process_id=1234, root_uri="file:///test", capabilities={})

    # Call initialize handler
    initialize(ls, params)

    # Check that settings remain default
    assert ls.settings["ruff_command"] is None

    # Reset after test
    reset_ruff_command()


def test_did_change_configuration_handler(ls):
    """Test that WORKSPACE_DID_CHANGE_CONFIGURATION handler updates settings."""
    from lsprotocol.types import DidChangeConfigurationParams
    from ruff_cgx import get_ruff_command, reset_ruff_command

    from collagraph_lsp.server import did_change_configuration

    # Reset to default first
    reset_ruff_command()

    # Create configuration change params
    params = DidChangeConfigurationParams(settings={"ruff_command": "updated-ruff"})

    # Call handler
    did_change_configuration(ls, params)

    # Check that settings were updated
    assert ls.settings["ruff_command"] == "updated-ruff"

    # Check that ruff command was applied
    assert get_ruff_command() == "updated-ruff"

    # Reset after test
    reset_ruff_command()



"""Semantic token provider for Collagraph .cgx files."""

import ast
import logging
import re
from typing import List

from collagraph.sfc.parser import CGXParser

logger = logging.getLogger(__name__)

# LSP semantic token types (standard types from the spec)
TOKEN_TYPES = [
    "namespace",
    "class",
    "enum",
    "interface",
    "struct",
    "typeParameter",
    "type",
    "parameter",
    "variable",
    "property",
    "enumMember",
    "decorator",
    "event",
    "function",
    "method",
    "macro",
    "label",
    "comment",
    "string",
    "keyword",
    "number",
    "regexp",
    "operator",
]

# LSP semantic token modifiers
TOKEN_MODIFIERS = [
    "declaration",
    "definition",
    "readonly",
    "static",
    "deprecated",
    "abstract",
    "async",
    "modification",
    "documentation",
    "defaultLibrary",
]


class SemanticToken:
    """Represents a semantic token with its position and type."""

    def __init__(
        self,
        line: int,
        start_char: int,
        length: int,
        token_type: str,
        modifiers: List[str] | None = None,
    ):
        self.line = line
        self.start_char = start_char
        self.length = length
        self.token_type = token_type
        self.modifiers = modifiers or []


class SemanticTokensProvider:
    """Provides semantic tokens for .cgx files."""

    def __init__(self):
        """Initialize the provider."""
        self.token_types = TOKEN_TYPES
        self.token_modifiers = TOKEN_MODIFIERS

    def get_tokens(
        self, content: str, file_path: str = "<stdin>"
    ) -> List[SemanticToken]:
        """
        Extract semantic tokens from a CGX file.

        Args:
            content: The CGX file content
            file_path: Optional file path for better error messages

        Returns:
            List of semantic tokens
        """
        tokens = []

        try:
            # Parse the CGX file
            parser = CGXParser()
            parser.feed(content)

            # Extract tokens from script section
            script_node = parser.root.child_with_tag("script")
            if script_node:
                script_tokens = self._extract_script_tokens(content, script_node)
                tokens.extend(script_tokens)

            # Extract tokens from template section
            template_nodes = [
                node
                for node in parser.root.children
                if not hasattr(node, "tag") or node.tag != "script"
            ]
            for template_node in template_nodes:
                template_tokens = self._extract_template_tokens(
                    content, template_node, parser
                )
                tokens.extend(template_tokens)

        except Exception as e:
            logger.error(f"Error extracting semantic tokens: {e}", exc_info=True)

        # Sort tokens by position
        tokens.sort(key=lambda t: (t.line, t.start_char))
        return tokens

    def _extract_script_tokens(self, content: str, script_node) -> List[SemanticToken]:
        """
        Extract semantic tokens from the script section using AST.

        Args:
            content: The full CGX content
            script_node: The parsed script node

        Returns:
            List of semantic tokens
        """
        tokens = []

        try:
            # Get script section line range
            start_line = script_node.location[0]
            end_line = script_node.end[0] - 1

            # Extract script content
            lines = content.splitlines()
            script_content = "\n".join(lines[start_line:end_line])

            # Parse the Python code
            tree = ast.parse(script_content)

            # Walk the AST and collect tokens
            for node in ast.walk(tree):
                if isinstance(node, ast.ClassDef):
                    # Class definition
                    tokens.append(
                        SemanticToken(
                            line=start_line + node.lineno - 1,
                            start_char=node.col_offset,
                            length=len(node.name),
                            token_type="class",
                            modifiers=["declaration"],
                        )
                    )

                elif isinstance(node, ast.FunctionDef):
                    # Function/method definition
                    token_type = "method" if self._is_method(node) else "function"
                    modifiers = ["declaration"]
                    if any(
                        isinstance(d, ast.Name) and d.id == "async"
                        for d in getattr(node, "decorator_list", [])
                    ):
                        modifiers.append("async")

                    tokens.append(
                        SemanticToken(
                            line=start_line + node.lineno - 1,
                            start_char=node.col_offset,
                            length=len(node.name),
                            token_type=token_type,
                            modifiers=modifiers,
                        )
                    )

                    # Parameters
                    for arg in node.args.args:
                        if arg.arg != "self" and arg.arg != "cls":
                            tokens.append(
                                SemanticToken(
                                    line=start_line + arg.lineno - 1,
                                    start_char=arg.col_offset,
                                    length=len(arg.arg),
                                    token_type="parameter",
                                    modifiers=["declaration"],
                                )
                            )

                elif isinstance(node, ast.Import):
                    # Import statements
                    for alias in node.names:
                        tokens.append(
                            SemanticToken(
                                line=start_line + node.lineno - 1,
                                start_char=node.col_offset,
                                length=len(alias.name),
                                token_type="namespace",
                            )
                        )

                elif isinstance(node, ast.ImportFrom):
                    # From import statements
                    if node.module:
                        tokens.append(
                            SemanticToken(
                                line=start_line + node.lineno - 1,
                                start_char=node.col_offset,
                                length=len(node.module),
                                token_type="namespace",
                            )
                        )

        except SyntaxError:
            # Script has syntax errors, skip token extraction
            pass
        except Exception as e:
            logger.debug(f"Error extracting script tokens: {e}")

        return tokens

    def _extract_template_tokens(
        self, content: str, template_node, parser
    ) -> List[SemanticToken]:
        """
        Extract semantic tokens from template section.

        Args:
            content: The full CGX content
            template_node: The parsed template node
            parser: The CGX parser

        Returns:
            List of semantic tokens
        """
        tokens = []

        try:
            lines = content.splitlines()

            # Recursively process template nodes
            self._process_template_node(template_node, lines, tokens)

        except Exception as e:
            logger.debug(f"Error extracting template tokens: {e}")

        return tokens

    def _process_template_node(
        self, node, lines: List[str], tokens: List[SemanticToken]
    ):
        """
        Recursively process a template node to extract tokens.

        Args:
            node: The template node to process
            lines: The source lines
            tokens: List to append tokens to
        """
        # Skip text and comment nodes
        if not hasattr(node, "tag"):
            return

        # Get the node's line
        if hasattr(node, "location"):
            line_num = node.location[0]
            if line_num < len(lines):
                line_content = lines[line_num]

                # Extract tag name
                tag_match = re.search(r"<(\w+)", line_content)
                if tag_match:
                    tag_name = tag_match.group(1)
                    tokens.append(
                        SemanticToken(
                            line=line_num,
                            start_char=tag_match.start(1),
                            length=len(tag_name),
                            token_type="class",  # Component names as classes
                        )
                    )

        # Process attributes
        if hasattr(node, "attrs"):
            for attr_name, attr_value in node.attrs.items():
                self._process_attribute(node, attr_name, attr_value, lines, tokens)

        # Process children
        if hasattr(node, "children"):
            for child in node.children:
                self._process_template_node(child, lines, tokens)

    def _process_attribute(
        self,
        node,
        attr_name: str,
        attr_value,
        lines: List[str],
        tokens: List[SemanticToken],
    ):
        """
        Process a template attribute and extract tokens from expressions.

        Args:
            node: The template node
            attr_name: The attribute name
            attr_value: The attribute value
            lines: The source lines
            tokens: List to append tokens to
        """
        if not hasattr(node, "location"):
            return

        line_num = node.location[0]
        if line_num >= len(lines):
            return

        # Find the attribute in the source
        # This is simplified - in production you'd need more robust parsing
        line_content = lines[line_num]

        # Highlight directive names (v-if, v-for, v-bind, v-on, etc.)
        if attr_name.startswith("v-"):
            attr_match = re.search(rf"\b({re.escape(attr_name)})\b", line_content)
            if attr_match:
                tokens.append(
                    SemanticToken(
                        line=line_num,
                        start_char=attr_match.start(1),
                        length=len(attr_name),
                        token_type="decorator",  # Directives as decorators
                    )
                )

        # Highlight shorthand bindings (: and @)
        elif attr_name.startswith(":") or attr_name.startswith("@"):
            attr_match = re.search(
                rf"([{attr_name[0]}]{re.escape(attr_name[1:])})", line_content
            )
            if attr_match:
                tokens.append(
                    SemanticToken(
                        line=line_num,
                        start_char=attr_match.start(1),
                        length=1,  # Just the : or @
                        token_type="operator",
                    )
                )
                # The property/event name
                tokens.append(
                    SemanticToken(
                        line=line_num,
                        start_char=attr_match.start(1) + 1,
                        length=len(attr_name) - 1,
                        token_type="event" if attr_name.startswith("@") else "property",
                    )
                )

        # Extract Python expressions from attribute values
        if isinstance(attr_value, str) and attr_value.strip():
            if attr_name.startswith((":", "@", "v-")):
                # This contains a Python expression
                # We could parse it, but for now just mark it as a variable reference
                # In production, you'd want to parse the expression properly
                pass

    def _is_method(self, node: ast.FunctionDef) -> bool:
        """
        Check if a function definition is a method (has self/cls parameter).

        Args:
            node: The function definition node

        Returns:
            True if it's a method
        """
        if node.args.args:
            first_arg = node.args.args[0].arg
            return first_arg in ("self", "cls")
        return False

    def encode_tokens(self, tokens: List[SemanticToken]) -> List[int]:
        """
        Encode tokens into the LSP semantic tokens format.

        The format is a flat array of integers where each token is represented by
        5 values:
        - Delta line (relative to previous token)
        - Delta start (relative to previous token if same line, absolute otherwise)
        - Length
        - Token type (index into token_types array)
        - Token modifiers (bit flags)

        Args:
            tokens: List of semantic tokens

        Returns:
            Encoded token data as list of integers
        """
        data = []
        prev_line = 0
        prev_start = 0

        for token in tokens:
            # Calculate deltas
            delta_line = token.line - prev_line
            delta_start = (
                token.start_char - prev_start if delta_line == 0 else token.start_char
            )

            # Get token type index
            try:
                token_type_idx = self.token_types.index(token.token_type)
            except ValueError:
                logger.warning(f"Unknown token type: {token.token_type}")
                token_type_idx = 0

            # Encode modifiers as bit flags
            modifier_bits = 0
            for modifier in token.modifiers:
                try:
                    modifier_idx = self.token_modifiers.index(modifier)
                    modifier_bits |= 1 << modifier_idx
                except ValueError:
                    logger.warning(f"Unknown token modifier: {modifier}")

            # Append the 5 values
            data.extend(
                [delta_line, delta_start, token.length, token_type_idx, modifier_bits]
            )

            # Update previous position
            prev_line = token.line
            prev_start = token.start_char

        return data



"""Main LSP server implementation for Collagraph files."""

import logging
from importlib.metadata import version

from lsprotocol.types import (
    INITIALIZE,
    TEXT_DOCUMENT_COMPLETION,
    TEXT_DOCUMENT_DID_CHANGE,
    TEXT_DOCUMENT_DID_CLOSE,
    TEXT_DOCUMENT_DID_OPEN,
    TEXT_DOCUMENT_DID_SAVE,
    TEXT_DOCUMENT_FORMATTING,
    TEXT_DOCUMENT_SEMANTIC_TOKENS_FULL,
    WORKSPACE_DID_CHANGE_CONFIGURATION,
    CompletionList,
    CompletionOptions,
    CompletionParams,
    DiagnosticSeverity,
    DidChangeConfigurationParams,
    DidChangeTextDocumentParams,
    DidCloseTextDocumentParams,
    DidOpenTextDocumentParams,
    DidSaveTextDocumentParams,
    DocumentFormattingParams,
    InitializeParams,
    Position,
    PublishDiagnosticsParams,
    Range,
    SemanticTokens,
    SemanticTokensLegend,
    SemanticTokensParams,
    TextEdit,
)
from lsprotocol.types import (
    Diagnostic as LspDiagnostic,
)
from pygls.lsp.server import LanguageServer
from ruff_cgx import (
    format_cgx_content,
    get_ruff_command,
    lint_cgx_content,
    reset_ruff_command,
    set_ruff_command,
)

from .completions import extract_script_region, get_python_completions
from .semantic_tokens import SemanticTokensProvider

logging.basicConfig(
    level=logging.INFO, format="%(asctime)s - %(name)s - %(levelname)s - %(message)s"
)
logger = logging.getLogger(__name__)


class CollagraphLanguageServer(LanguageServer):
    """Language server for Collagraph .cgx files."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.semantic_tokens_provider = SemanticTokensProvider()
        # Configuration settings
        self.settings = {
            "ruff_command": None,  # None means use ruff_cgx default (env var or "ruff")
        }


# Create the server instance
server = CollagraphLanguageServer("collagraph-lsp", f"v{version('collagraph_lsp')}")


def _apply_ruff_command(ruff_command: str | None):
    """
    Apply the ruff command configuration.

    Args:
        ruff_command: Path or command name for ruff executable, or None to use default
    """
    if ruff_command:
        logger.info(f"Setting ruff command to: {ruff_command}")
        set_ruff_command(ruff_command)
    else:
        # Use ruff_cgx default (env var or "ruff")
        reset_ruff_command()
        logger.info(f"Using default ruff command: {get_ruff_command()}")


def _severity_to_lsp(severity: str) -> DiagnosticSeverity:
    """Convert our severity string to LSP DiagnosticSeverity."""
    mapping = {
        "error": DiagnosticSeverity.Error,
        "warning": DiagnosticSeverity.Warning,
        "info": DiagnosticSeverity.Information,
        "hint": DiagnosticSeverity.Hint,
    }
    return mapping.get(severity, DiagnosticSeverity.Warning)


def validate_document(ls: CollagraphLanguageServer, uri: str):
    """
    Validate a CGX document and publish diagnostics.

    Args:
        ls: The language server instance
        uri: The document URI
    """
    try:
        # Only validate .cgx files
        if not uri.endswith(".cgx"):
            logger.debug(f"Skipping non-CGX file: {uri}")
            return

        # Get the document (pygls 2.0 uses get_text_document)
        doc = ls.workspace.get_text_document(uri)
        content = doc.source

        logger.info(f"Validating document: {uri}")

        # Run the linter
        diagnostics = lint_cgx_content(content)

        # Convert to LSP diagnostics
        lsp_diagnostics = []
        for diag in diagnostics:
            lsp_diag = LspDiagnostic(
                range=Range(
                    start=Position(line=diag.line, character=diag.column),
                    end=Position(line=diag.end_line, character=diag.end_column),
                ),
                message=diag.message,
                severity=_severity_to_lsp(diag.severity),
                code=diag.code,
                source=diag.source,
            )
            lsp_diagnostics.append(lsp_diag)

        # Publish diagnostics (pygls 2.0 uses PublishDiagnosticsParams)
        params = PublishDiagnosticsParams(uri=uri, diagnostics=lsp_diagnostics)
        ls.text_document_publish_diagnostics(params)
        logger.info(f"Published {len(lsp_diagnostics)} diagnostics for {uri}")

    except Exception as e:
        logger.error(f"Error validating document {uri}: {e}", exc_info=True)


@server.feature(INITIALIZE)
def initialize(ls: CollagraphLanguageServer, params: InitializeParams):
    """
    Handle initialization request from the client.

    Accepts configuration via initialization options:
    - ruff_command: Path or command name for ruff executable

    Example initialization options from client:
    {
        "ruff_command": "/opt/homebrew/bin/ruff"
    }
    """
    logger.info("Initializing Collagraph LSP server")

    # Get initialization options
    init_options = params.initialization_options or {}

    # Extract settings
    ruff_command = init_options.get("ruff_command")

    # Store settings
    if ruff_command is not None:
        ls.settings["ruff_command"] = ruff_command

    # Apply ruff command configuration
    _apply_ruff_command(ls.settings["ruff_command"])

    logger.info(f"Server initialized with settings: {ls.settings}")


@server.feature(WORKSPACE_DID_CHANGE_CONFIGURATION)
def did_change_configuration(
    ls: CollagraphLanguageServer, params: DidChangeConfigurationParams
):
    """
    Handle configuration change notifications from the client.

    Expects settings in the format:
    {
        "ruff_command": "/opt/homebrew/bin/ruff"
    }
    """
    logger.info("Configuration changed")

    # Get settings from params
    settings = params.settings

    # Extract collagraph-lsp settings
    if settings and "ruff_command" in settings:
        # Update ruff command if provided
        ls.settings["ruff_command"] = settings["ruff_command"]
        _apply_ruff_command(ls.settings["ruff_command"])
        logger.info(f"Updated settings: {ls.settings}")


@server.feature(TEXT_DOCUMENT_DID_OPEN)
def did_open(ls: CollagraphLanguageServer, params: DidOpenTextDocumentParams):
    """Handle document open event."""
    logger.info(f"Document opened: {params.text_document.uri}")
    validate_document(ls, params.text_document.uri)


@server.feature(TEXT_DOCUMENT_DID_CHANGE)
def did_change(ls: CollagraphLanguageServer, params: DidChangeTextDocumentParams):
    """Handle document change event."""
    logger.info(f"Document changed: {params.text_document.uri}")
    validate_document(ls, params.text_document.uri)


@server.feature(TEXT_DOCUMENT_DID_SAVE)
def did_save(ls: CollagraphLanguageServer, params: DidSaveTextDocumentParams):
    """Handle document save event."""
    logger.info(f"Document saved: {params.text_document.uri}")
    validate_document(ls, params.text_document.uri)


@server.feature(TEXT_DOCUMENT_DID_CLOSE)
def did_close(ls: CollagraphLanguageServer, params: DidCloseTextDocumentParams):
    """Handle document close event."""
    logger.info(f"Document closed: {params.text_document.uri}")
    # Clear diagnostics for closed document
    clear_params = PublishDiagnosticsParams(
        uri=params.text_document.uri, diagnostics=[]
    )
    ls.text_document_publish_diagnostics(clear_params)


@server.feature(TEXT_DOCUMENT_FORMATTING)
def formatting(ls: CollagraphLanguageServer, params: DocumentFormattingParams):
    """Handle document formatting request."""
    uri = params.text_document.uri
    logger.info(f"Formatting document: {uri}")

    try:
        # Only format .cgx files
        if not uri.endswith(".cgx"):
            logger.debug(f"Skipping non-CGX file: {uri}")
            return []

        # Get the document
        doc = ls.workspace.get_text_document(uri)
        content = doc.source

        # Format the content
        formatted_content = format_cgx_content(content, uri)

        # If content didn't change, return empty list
        if formatted_content == content:
            logger.info(f"No formatting changes needed for {uri}")
            return []

        # Calculate the range that needs to be replaced (entire document)
        lines = content.splitlines(keepends=True)
        last_line = len(lines) - 1
        last_char = len(lines[-1]) if lines else 0

        # Create a text edit that replaces the entire document
        text_edit = TextEdit(
            range=Range(
                start=Position(line=0, character=0),
                end=Position(line=last_line, character=last_char),
            ),
            new_text=formatted_content,
        )

        logger.info(f"Formatted document {uri}")
        return [text_edit]

    except Exception as e:
        logger.error(f"Error formatting document {uri}: {e}", exc_info=True)
        return []


@server.feature(
    TEXT_DOCUMENT_SEMANTIC_TOKENS_FULL,
    SemanticTokensLegend(
        token_types=SemanticTokensProvider().token_types,
        token_modifiers=SemanticTokensProvider().token_modifiers,
    ),
)
def semantic_tokens_full(
    ls: CollagraphLanguageServer, params: SemanticTokensParams
) -> SemanticTokens:
    """
    Handle semantic tokens request.

    Provides rich syntax highlighting based on semantic analysis of the code.
    """
    uri = params.text_document.uri
    logger.info(f"Providing semantic tokens for: {uri}")

    try:
        # Only process .cgx files
        if not uri.endswith(".cgx"):
            logger.debug(f"Skipping non-CGX file: {uri}")
            return SemanticTokens(data=[])

        # Get the document
        doc = ls.workspace.get_text_document(uri)
        content = doc.source

        # Extract semantic tokens
        tokens = ls.semantic_tokens_provider.get_tokens(content, uri)

        # Encode tokens in LSP format
        encoded_data = ls.semantic_tokens_provider.encode_tokens(tokens)

        logger.info(f"Provided {len(tokens)} semantic tokens for {uri}")
        return SemanticTokens(data=encoded_data)

    except Exception as e:
        logger.error(f"Error providing semantic tokens for {uri}: {e}", exc_info=True)
        return SemanticTokens(data=[])


@server.feature(
    TEXT_DOCUMENT_COMPLETION,
    CompletionOptions(
        trigger_characters=[".", "[", "(", '"', "'"],
        resolve_provider=False,
    ),
)
async def completions(
    ls: CollagraphLanguageServer, params: CompletionParams
) -> CompletionList:
    """
    Handle completion request.

    Phase 1: Provides Python completions in <script> sections only.
    """
    uri = params.text_document.uri
    position = params.position
    logger.info(
        f"Providing completions for: {uri} at {position.line}:{position.character}"
    )

    try:
        # Only process .cgx files
        if not uri.endswith(".cgx"):
            logger.debug(f"Skipping non-CGX file: {uri}")
            return CompletionList(is_incomplete=False, items=[])

        # Get the document
        doc = ls.workspace.get_text_document(uri)
        content = doc.source

        # Extract script region at the cursor position
        script_region = extract_script_region(content, position)

        # Phase 1: Only provide completions for <script> sections
        if script_region.in_script:
            items = await get_python_completions(script_region)
            logger.info(f"Provided {len(items)} completions for {uri}")
        else:
            items = []
            logger.debug(
                "Cursor not in script section at "
                f"{uri}:{position.line}:{position.character}"
            )

        return CompletionList(is_incomplete=False, items=items)

    except Exception as e:
        logger.error(f"Error providing completions for {uri}: {e}", exc_info=True)
        return CompletionList(is_incomplete=False, items=[])


def main():
    """Main entry point for the LSP server."""

    # Start the server using stdin/stdout
    logger.info("Starting Collagraph LSP server...")
    server.start_io()


if __name__ == "__main__":
    main()



"""Completion support for Collagraph .cgx files."""

import logging
import re
from dataclasses import dataclass

import jedi
from lsprotocol.types import (
    CompletionItem,
    CompletionItemKind,
    InsertTextFormat,
    Position,
)


@dataclass
class ScriptRegion:
    """Information about a script region and the cursor position within it."""

    in_script: bool
    script_content: str | None
    script_line: int | None  # Line within script content (0-based)
    script_column: int | None  # Column within script content (0-based)


def extract_script_region(source: str, position: Position) -> ScriptRegion:
    """
    Extract script region and transform position to be relative to script content.

    Handles multiple <script> sections by finding the one containing the cursor.

    Returns:
        ScriptRegion with:
        - in_script: True if cursor is inside <script> content
        - script_content: The Python code inside <script> tags (if in_script)
        - script_line: Line number within script content (0-based, if in_script)
        - script_column: Column number within script content (0-based, if in_script)
    """
    script_pattern = r"<script[^>]*>(.*?)</script>"

    # Calculate offset of cursor position in full source
    source_offset = position_to_offset(source, position)

    # Find all script sections and check which one contains the cursor
    for match in re.finditer(script_pattern, source, re.DOTALL):
        script_content = match.group(1)
        script_start_offset = match.start(1)
        script_end_offset = match.end(1)

        # Check if cursor is within this script content
        if script_start_offset <= source_offset <= script_end_offset:
            # Calculate position within script content
            script_offset = source_offset - script_start_offset

            # Convert offset to line/column within script
            script_lines = script_content.split("\n")
            current_offset = 0
            script_line = 0
            script_column = 0

            for line_idx, line in enumerate(script_lines):
                if current_offset + len(line) + 1 > script_offset:  # +1 for newline
                    script_line = line_idx
                    script_column = script_offset - current_offset
                    break
                current_offset += len(line) + 1

            return ScriptRegion(
                in_script=True,
                script_content=script_content,
                script_line=script_line,
                script_column=script_column,
            )

    # No script section contains the cursor
    return ScriptRegion(
        in_script=False,
        script_content=None,
        script_line=None,
        script_column=None,
    )


def position_to_offset(source: str, position: Position) -> int:
    """Convert Position (line, character) to string offset."""
    lines = source.split("\n")
    offset = sum(len(line) + 1 for line in lines[: position.line])  # +1 for newline
    offset += position.character
    return offset


async def get_python_completions(
    script_region: ScriptRegion,
) -> list[CompletionItem]:
    """
    Get Python completions using Jedi for script sections.

    Phase 1: Provide completions for <script> section content.
    Jedi can handle incomplete code (like "os.") which is common during typing.

    Args:
        script_region: ScriptRegion containing script content and cursor position

    Returns:
        List of completion items from Jedi
    """
    if not script_region.in_script or not script_region.script_content:
        return []

    try:
        # Use Jedi for Python completions on the script content
        # Jedi handles incomplete code gracefully
        script = jedi.Script(code=script_region.script_content)

        # Get completions (Jedi uses 1-based line numbers)
        completions = script.complete(
            line=script_region.script_line + 1,
            column=script_region.script_column,
        )

        # Convert Jedi completions to LSP CompletionItems
        items = []
        for comp in completions:
            items.append(
                CompletionItem(
                    label=comp.name,
                    kind=map_jedi_type_to_lsp(comp.type),
                    detail=comp.description,
                    documentation=comp.docstring(raw=True)
                    if comp.docstring()
                    else None,
                    insert_text=comp.name,
                    insert_text_format=InsertTextFormat.PlainText,
                    sort_text=comp.name,
                )
            )

        return items

    except Exception as e:
        # Return empty list on any error
        logging.warning(f"Completion failed with error: {e}")
        return []


def map_jedi_type_to_lsp(jedi_type: str) -> CompletionItemKind:
    """Map Jedi completion types to LSP CompletionItemKind."""
    mapping = {
        "module": CompletionItemKind.Module,
        "class": CompletionItemKind.Class,
        "function": CompletionItemKind.Function,
        "param": CompletionItemKind.Variable,
        "path": CompletionItemKind.File,
        "keyword": CompletionItemKind.Keyword,
        "property": CompletionItemKind.Property,
        "statement": CompletionItemKind.Variable,
    }
    return mapping.get(jedi_type, CompletionItemKind.Text)



"""Tests for completion support in Collagraph .cgx files.

Phase 1: Only tests completions within <script> tags.
"""

from textwrap import dedent

import pytest
from lsprotocol.types import Position

from collagraph_lsp.completions import (
    extract_script_region,
    get_python_completions,
    position_to_offset,
)


class TestContextDetection:
    """Test detection of Python regions in .cgx files."""

    def test_detect_script_region(self):
        source = dedent(
            """
            <template>
              <div>Hello</div>
            </template>
            <script>
            import os
            x = 10
            </script>"""
        ).lstrip()

        # Cursor in script section
        position = Position(line=4, character=7)  # "import"
        assert extract_script_region(source, position).in_script

        # Cursor outside script section
        position = Position(line=1, character=5)  # in template
        assert not extract_script_region(source, position).in_script

    def test_extract_script_section(self):
        """Test that extract_script_region correctly identifies script sections."""
        source = dedent(
            """
            <template>
              <div>Hello</div>
            </template>
            <script>
            import os
            x = 10
            </script>"""
        ).lstrip()

        script_region = extract_script_region(source, Position(line=4, character=7))
        assert script_region.in_script
        assert script_region.script_content is not None
        assert "import os" in script_region.script_content

    def test_extract_template_region(self):
        """Test that extract_script_region correctly identifies non-script regions."""
        source = dedent(
            """
            <template>
              <div>Hello World</div>
            </template>"""
        ).lstrip()

        script_region = extract_script_region(source, Position(line=1, character=5))
        assert not script_region.in_script
        assert script_region.script_content is None


class TestPythonCompletions:
    """Test Python completions in script sections."""

    def test_completion_context_in_script(self):
        """Test that script region is correctly extracted for completions."""
        source = dedent(
            """
            <template>
              <div>Hello</div>
            </template>
            <script>
            import os
            os.
            </script>"""
        ).lstrip()

        script_region = extract_script_region(source, Position(line=5, character=3))
        assert script_region.in_script
        assert script_region.script_content is not None

    @pytest.mark.asyncio
    async def test_jedi_test(self):
        """Test that Jedi returns exactly the expected completions for string
        methods."""
        source = dedent(
            """




            <script>
            toet = "asdf"
            toet.st
            </script>"""
        )

        # Position after "toet.st" - line 7 in the full file
        # (after opening newline + 4 blank lines + <script> + toet = "asdf")
        position = Position(line=7, character=7)
        script_region = extract_script_region(source, position)
        assert script_region.in_script

        # Get actual completions from Jedi
        items = await get_python_completions(script_region)

        # Should have completions for string methods starting with "st"
        # e.g., startswith, strip, etc.
        assert len(items) >= 2, f"Expected at least 2 completions, got {len(items)}"
        labels = [item.label for item in items]
        assert any("start" in label.lower() for label in labels), (
            f"Expected string methods starting with 'st', got: {labels}"
        )

    @pytest.mark.asyncio
    async def test_actual_jedi_completions_returned(self):
        """Test that Jedi actually returns completion items for Python code."""
        source = dedent(
            """
            <template>
              <div>Hello</div>
            </template>
            <script>
            import os
            os.
            </script>"""
        ).lstrip()

        # Position after "os." on line 5 (0-indexed)
        position = Position(line=5, character=3)
        script_region = extract_script_region(source, position)

        # Get actual completions from Jedi
        items = await get_python_completions(script_region)

        # Should have completions for os module attributes
        assert len(items) > 0, "Expected Jedi to return completions for 'os.'"

        # Check that items have proper LSP structure
        assert all(hasattr(item, "label") for item in items)
        assert all(hasattr(item, "kind") for item in items)

        # Should include common os module attributes
        labels = [item.label for item in items]
        assert any("path" in label for label in labels), (
            "Expected 'path' in os completions"
        )

    @pytest.mark.asyncio
    async def test_completions_for_builtin_module(self):
        """Test completions for builtin module attributes."""
        source = dedent(
            """
            <script>
            import sys
            sys.
            </script>
            <template>
              <div>Test</div>
            </template>"""
        ).lstrip()

        # Position after "sys."
        position = Position(line=2, character=4)
        script_region = extract_script_region(source, position)

        items = await get_python_completions(script_region)

        # Should have completions for sys module
        assert len(items) > 0, "Expected completions for sys module"

        labels = [item.label for item in items]
        # Check for common sys attributes
        assert any("version" in label or "path" in label for label in labels), (
            "Expected sys module attributes in completions"
        )

    @pytest.mark.asyncio
    async def test_completions_with_component_context(self):
        """Test that completions work for component classes."""
        source = dedent(
            """
            <script>
            from collagraph import Component

            class MyComponent(Component):
                def __init__(self):
                    self.my_value = 42

                def my_method(self):
                    self.
            </script>
            <template>
              <div>{{ my_value }}</div>
            </template>"""
        ).lstrip()

        # Position after "self."
        position = Position(line=8, character=13)
        script_region = extract_script_region(source, position)

        items = await get_python_completions(script_region)

        # Should have completions for self attributes
        assert len(items) > 0, "Expected completions for 'self.'"

        labels = [item.label for item in items]
        # Should see our custom attributes/methods
        assert any("my_value" in label or "my_method" in label for label in labels), (
            "Expected component attributes in completions"
        )

    def test_no_completion_in_template(self):
        """Test that completions are not provided in template regions."""
        source = dedent(
            """
            <template>
              <div>Hello World</div>
            </template>
            <script>
            import os
            </script>"""
        ).lstrip()

        # Cursor in template section
        script_region = extract_script_region(source, Position(line=1, character=5))
        assert not script_region.in_script

    @pytest.mark.asyncio
    async def test_completions_with_invalid_syntax(self):
        """Test that completions handle invalid syntax gracefully."""
        # Invalid Python syntax
        source = dedent(
            """
            <script>
            this is not valid python syntax!!!
            </script>
            <template>
              <div>Test</div>
            </template>"""
        ).lstrip()

        position = Position(line=1, character=10)
        script_region = extract_script_region(source, position)

        # Should not crash - Jedi may provide keyword completions
        items = await get_python_completions(script_region)
        # Jedi is robust and may provide some completions even with invalid syntax
        assert isinstance(items, list), "Expected list of completions (possibly empty)"


class TestEdgeCases:
    """Test edge cases and error handling for script section detection."""

    def test_malformed_cgx_file(self):
        """Test handling of malformed .cgx files."""
        source = "<script>import os\nos."  # No closing tag

        # Should not crash
        script_region = extract_script_region(source, Position(line=1, character=3))
        # Without closing tag, won't detect as script region (acceptable limitation)
        # The important thing is it doesn't crash
        assert isinstance(script_region.in_script, bool)

    def test_script_with_type_attribute(self):
        """Test <script> tag with type attribute."""
        source = dedent(
            """
            <script type="text/python">
            import os
            os.path
            </script>"""
        ).lstrip()

        position = Position(line=2, character=3)  # "os.path"
        assert extract_script_region(source, position).in_script

    def test_multiple_script_sections(self):
        """Test handling of multiple <script> sections."""
        source = dedent(
            """
            <script>
            import os
            </script>
            <template>
                <div>Test</div>
            </template>
            <script>
            import sys
            </script>"""
        ).lstrip()

        # First script section
        assert extract_script_region(source, Position(line=1, character=7)).in_script
        # Second script section
        assert extract_script_region(source, Position(line=7, character=7)).in_script
        # Template section (between scripts)
        assert not extract_script_region(
            source, Position(line=4, character=5)
        ).in_script

    def test_position_to_offset_multiline(self):
        """Test position_to_offset helper with multiline content."""
        source = dedent(
            """
            line1
            line2
            line3"""
        ).lstrip()

        # Start of file
        assert position_to_offset(source, Position(line=0, character=0)) == 0
        # Start of second line
        assert position_to_offset(source, Position(line=1, character=0)) == 6
        # Middle of second line
        assert position_to_offset(source, Position(line=1, character=3)) == 9

    def test_cursor_on_script_tag_boundary(self):
        """Test cursor position at the boundaries of script tag."""
        source = dedent(
            """
            <script>
            import os
            </script>
            """
        ).lstrip()

        # Just inside opening tag
        assert extract_script_region(source, Position(line=1, character=0)).in_script
        # Just before closing tag
        assert extract_script_region(source, Position(line=1, character=9)).in_script
        # On the opening tag itself - should not be inside
        assert not extract_script_region(
            source, Position(line=0, character=5)
        ).in_script
